from kadena_sdk import hash_and_sign, blake_hash
from kadena_sdk import KadenaSdk


def test_sign_message():
    # 09dbbe2d2caf283a5da170cc438eed631b5898b921512cae8720232313483a3e
    private_key = '1de84cf16631a778317e1c33a6b729875734c129b0094e809713b7225fd3dfb7'
    message = "Hello, World!"
    hash_code, sig = hash_and_sign(message, private_key)
    assert(hash_code == 'URvIHd4RGAg4xWLIK7NfMiP0YGHr3kqVXCez9InPHgM')
    assert(sig == '756ec1891523f56369c98bf2544a2f91033de9cea53773d7cc4d7812babd9a1d0ab03b706e628bd1636fb5d65083711ba6584bf6ffe471a48657e6a571031306')

    print(hash_code, sig)


def test_hash():
  assert(blake_hash("hello") == "Mk3PAn3UowqTLEQfNlol6GsXPe-kuOWJSCU0cbgbcs8")
  assert(blake_hash({ "foo": {"int":1} }) == "h9BZgylRf_M4HxcBXr15IcSXXXSz74ZC2IAViGle_z4")


from datetime import datetime
import time
import json
import requests

from kadena_sdk.crypto_lib import *
from kadena_sdk.key_pair import KeyPair
from kadena_sdk.types import *

class KadenaSdk():

  SEND = '/send'
  LOCAL = '/local'
  LISTEN = '/listen'

  def __init__(self, base_url: str='https://api.testnet.chainweb.com', key_pair: KeyPair=None):
    self.base_url = base_url
    self.network_id = requests.get(f'{base_url}/config').json()['chainwebVersion']
    self.key_pair = key_pair


  def build_exec_payload(self, pact_code: str, env_data: dict={}) -> ExecPayload:
    """Builds an execute payload object for use with build_command."""
    return {
      'exec': {
        'code': pact_code,
        'data': env_data
      }
    }


  def build_cont_payload(self, 
    pact_id: str, 
    step: int,
    rollback: bool, 
    proof: str,
    env_data: dict={}) -> ContPayload:
    """Builds an execute payload object for use with build_command."""
    return {
      'cont': {
        'pactId': pact_id,
        'step': step,
        'rollback': rollback,
        'data': env_data,
        'proof': proof,
      }
    }


  def build_command(self, 
    payload: ExecPayload | ContPayload,
    chain_ids: list[str],
    signers: list[str]=None, 
    sender: str=None, 
    gas_price: float=1.0e-5, 
    gas_limit: int=2500):
    """Build a command that can be committed (send) to or locally (local)
    sent to the Kadena network.
    Will create a command for each chain_id in the chain_ids list.
    
    Payload must be in the format:
    {
      "exec": {
        "data": dict,
        "code": str
      }  
    }
    """
    # Create Time Stamp
    t_epoch = time.time()
    t_epoch = round(t_epoch) - 15

    sender_actual = sender
    if sender_actual == None:
      if self.key_pair:
        sender_actual = f'k:{self.key_pair.get_pub_key()}'
      else:
        # If we have no sender, just use an empty sender.
        sender_actual = ''

    print(signers)
    signers_actual = signers
    print(signers)
    print(self.key_pair)
    if signers == None:
      if self.key_pair:
        signers_actual = [{ "pubKey": self.key_pair.get_pub_key() }]
      else:
        signers_actual = []
    print(signers_actual)

    cmds = {}
    for chain_id in chain_ids:
      command = {
        "networkId": self.network_id,
        "payload": payload,
        "signers": signers_actual,
        "meta": {
          "gasLimit": gas_limit,
          "chainId": chain_id,
          "gasPrice": gas_price,
          "sender": sender_actual,
          "ttl": 28000,
          "creationTime": t_epoch
        },
        "nonce": datetime.now().strftime("%Y%m%d%H%M%S")
      }
      cmds[chain_id] = command
      # cmds.append(command)

    return cmds


  def send(self, cmds: dict):
    """Commits a list of commands to the Kadena network.
    Each command is an object created via build_command.
    Uses the chainId from the command to determine which chain to commit to.
    
    cmds must be in the format:
    {
      "0": command,
      "1": command,
      ...
    }
    """
    ret = {}

    for command in cmds.values():
      cmd_json = json.dumps(command)
      hash_code, sig = self.hash_and_sign(cmd_json)
    
      to_post = {
        'cmds': {
          'hash': hash_code,
          'sigs': [{'sig': sig}],
          'cmd': cmd_json,
        }
      }

      chain_id = command['meta']['chainId']
      ret[chain_id] = requests.post(
        self.build_url(self.SEND, chain_id), 
        json=to_post)

    return ret
  
  
  def local(self, cmds: dict):
    """Locally (dirty) reads from Kadena using a list of commands.
    Each command is an object created via build_command.
    Uses the chainId from the command to determine which chain to read from.
    
    cmds must be in the format:
    {
      "0": command,
      "1": command,
      ...
    }
    """
    ret = {}

    for command in cmds.values():
      cmd_json = json.dumps(command)
      hash_code, sig = self.hash_and_sign(cmd_json)

      sigs = [] if sig is None else [{'sig': sig}]
    
      to_post = {
        'hash': hash_code,
        'sigs': sigs,
        'cmd': cmd_json,
      }

      chain_id = command['meta']['chainId']
      ret[chain_id] = requests.post(
        self.build_url(self.LOCAL, command['meta']['chainId']), 
        json=to_post)

    return ret
  

  def run_pact(self, pact_code: str, env_data: dict={}, chain_id: str='0'):
    """Runs a pact code string on the Kadena network.
    Returns a dict with the results of the local run."""
    
    payload = self.build_exec_payload(pact_code, env_data)

    return self.local(self.build_command(payload, ['0']))['0'].json()['result']


  def listen(self, tx_id: str, chain_id: list):
    """Blocking. Listens to a transaction on the given chain."""
    data = {
      'listen': tx_id
    }

    return requests.post(self.build_url(self.LISTEN, chain_id), json=data)
  

  def send_and_listen(self, command):
    """Commits a command to the Kadena network and listens to the transaction.
    Uses the chainId from the command to determine which chain to listen on."""
    chain_id = command['meta']['chainId']
    result = self.send({ chain_id: command })
    tx_id = result.json()['requestKeys'][0]
    return self.listen(tx_id, chain_id)
  

  def build_url(self, endpoint, chain_id):
    """Builds a url for a given endpoint and chain_id"""
    url = f'{self.base_url}/chainweb/0.0/{self.network_id}/chain/{chain_id}/pact/api/v1{endpoint}'
    return url


  def hash_and_sign(self, command_json):
    """Signs a command with the private key of the locally stored 
    KeyPair instance."""
    bhash = blake_hash(command_json)
    sig = None

    if self.key_pair:
      sig = sign(bhash, self.key_pair.get_priv_key())

    return bhash, sig
    
  

  def set_key_pair(self, key_pair: KeyPair):
    self.key_pair = key_pair


import pytest

from kadena_sdk import KadenaSdk, KeyPair

@pytest.fixture
def key_pair():
  key_pair = KeyPair(type='json', 
    priv_key='5e8b125c89ed409f2cfcc6e863e8aafd60b9d80a4d2333a12592f7a961a62bf8',
    pub_key='ad273a54460305767e2e36f41d1a5fe78c48474a6e3bc18624d53fbbbb5974bb')
  return key_pair

@pytest.fixture
def kadena(key_pair):
  kadena = KadenaSdk(key_pair=key_pair)
  return kadena

@pytest.fixture
def kadena_no_keys():
  kadena_no_keys = KadenaSdk()
  return kadena_no_keys

@pytest.fixture
def payload_hello():
  payload = {
    "exec": {
      "data": {},
      "code": '(format "Test {}" ["hello"])'
    }
  }
  return payload

def test_keypair(key_pair):
  assert(key_pair.get_pub_key() == 'ad273a54460305767e2e36f41d1a5fe78c48474a6e3bc18624d53fbbbb5974bb')
  assert(key_pair.get_priv_key() == '5e8b125c89ed409f2cfcc6e863e8aafd60b9d80a4d2333a12592f7a961a62bf8')

def test_build_command(kadena: KadenaSdk, payload_hello: dict):
  chain_ids = ['0', '1']
  cmds = kadena.build_command(payload_hello, chain_ids)

  for c in chain_ids:
    assert(cmds[c]['networkId'] == 'testnet04')
    assert(cmds[c]['payload'] == payload_hello)
    assert(cmds[c]['signers'][0]['pubKey'] == 'ad273a54460305767e2e36f41d1a5fe78c48474a6e3bc18624d53fbbbb5974bb')
    assert(cmds[c]['meta']['gasLimit'] == 2500)
    assert(cmds[c]['meta']['chainId'] == c)
    assert(cmds[c]['meta']['gasPrice'] == 1.0e-5)
    assert(cmds[c]['meta']['sender'] == 'k:ad273a54460305767e2e36f41d1a5fe78c48474a6e3bc18624d53fbbbb5974bb')
    assert(cmds[c]['meta']['ttl'] == 28000)

def test_local(kadena: KadenaSdk, kadena_no_keys: KadenaSdk, payload_hello: dict):
  print('Local with keys')
  chain_ids = ['0', '1']
  cmds = kadena.build_command(payload_hello, chain_ids)
  res = kadena.local(cmds)
  assert(res['0'].json()['result']['data'] == 'Test hello')
  assert(res['1'].json()['result']['data'] == 'Test hello')

  print('Local without keys')
  cmds = kadena_no_keys.build_command(payload_hello, chain_ids)
  res = kadena_no_keys.local(cmds)
  assert(kadena_no_keys.key_pair is None)
  print(res['0'].text)
  assert(res['0'].json()['result']['data'] == 'Test hello')
  assert(res['1'].json()['result']['data'] == 'Test hello')

def text_run_pact(kadena_no_keys: KadenaSdk):
  res = kadena_no_keys.run_pact(
    '(format "Test {}" [(read-msg "test")])', 
    env_data={'test': 'hello'}
  )
  assert(res == 'Test hello')



from typing import TypedDict

class ExecMessage(TypedDict):
  code: str
  data: dict

class ContMessage(TypedDict):
  pactId: str
  step: int
  rollback: bool
  data: dict
  proof: str

class ExecPayload(TypedDict):
  exec: ExecMessage

class ContPayload(TypedDict):
  cont: ContMessage




import json


class KeyPair():

  def __init__(self, key_file='', type='keyring', **kwargs):
    if type == 'keyring':
      with open(key_file, 'r') as f:
        import keyring
        text = f.read()
        jason = json.loads(text)
        priv_key_info = jason['priv_key']
        self.priv_key = keyring.get_password(priv_key_info['service'], priv_key_info['username'])
        self.pub_key = jason['pub_key']
    elif type == 'json':
      self.priv_key = kwargs['priv_key']
      self.pub_key = kwargs['pub_key']
    elif type == 'pub_only':
      self.priv_key = ''
      self.pub_key = kwargs['pub_key']
  

  def get_priv_key(self):
    return self.priv_key
  

  def get_pub_key(self):
    return self.pub_key


from nacl.signing import SigningKey
import base64
import hashlib
import json

def hash_and_sign(s, priv_key):
  # Turn the json into a byte string
  cmd_data = bytes(s, encoding="utf8")

  # Create the signature
  sk = SigningKey(bytes.fromhex(priv_key))

  # Create the hash code
  hash_bytes = blake2b(cmd_data)
  hash_code = base64.urlsafe_b64encode(hash_bytes).decode().rstrip('=')
  
  return hash_code, sk.sign(hash_bytes).signature.hex()


def sign(hash, priv_key):
  """Accepts a blake hash and a private key and returns a signature."""
  sk = SigningKey(bytes.fromhex(priv_key))

  # Get the bytes from the hash
  hash_bytes = base64.urlsafe_b64decode(f'{hash}==')

  return sk.sign(hash_bytes).signature.hex()


def blake_hash(value):
  """Blake2b hashes a string or dict."""
  
  if type(value) is str:
    hash_bytes = blake2b(bytes(value, encoding="utf8"))
    return base64.urlsafe_b64encode(hash_bytes).decode().rstrip('=')
  elif type(value) is dict:
    s = json.dumps(value, separators=(',', ':'))
    hash_bytes = blake2b(bytes(s, encoding="utf8"))
    return base64.urlsafe_b64encode(hash_bytes).decode().rstrip('=')

  raise Exception(f'Invalid type: {type(value)}')

def blake2b(bytes, digest_size=32):
  hash2b = hashlib.blake2b(bytes, digest_size=digest_size)
  return hash2b.digest()


